/*
-----------------------------------------------------------------------------
This file is a part of Gsage engine

Copyright (c) 2014-2016 Artem Chernyshev

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
-----------------------------------------------------------------------------
*/

#ifndef _Engine_H_
#define _Engine_H_

#define KEY_ID "id"
#define KEY_FLAGS "flags"

#include "EventDispatcher.h"
#include "GsageDefinitions.h"
#include "Entity.h"
#include "EngineSystem.h"
#include "EngineEvent.h"

#include "ObjectPool.h"
#include <map>
#include <vector>
#include <thread>
#include <atomic>

#include "DataProxy.h"
#include "ThreadSafeQueue.h"

namespace Gsage
{
  class EntityComponent;
  class EngineSystem;

  /**
   * Core class of the game engine
   */
  class Engine : public EventDispatcher
  {
    public:
      typedef std::map<std::string, EngineSystem*> EngineSystems;
      typedef ObjectPool<Entity> Entities;

      typedef std::function<bool(Entity*)> UnloadMatcherFunc;

      Engine(const unsigned int& poolSize = ENTITY_POOL_SIZE);
      virtual ~Engine();
      /**
       * Initializes all the systems in the order in which it were added
       * @param configuration DataProxy with engine configuration
       * @param environment global application environment
       * @returns false if fails to set up one of them
       */
      bool initialize(const DataProxy& configuration, const DataProxy& environment);
      /**
       * Configure all systems in the engine
       * @param configuration DataProxy with system configs
       * @returns false if fails to configure one of the systems
       */
      bool configureSystems(const DataProxy& config);
      /**
       * Updates each system
       * @param time Delta time
       */
      void update(const double& time);
      /**
       * Add system to the engine
       * @param configure system after adding
       * @returns system pointer if succeed
       */
      template<typename T>
      T* addSystem(bool configure = true)
      {
        T* system = new T();
        if(addSystem(T::type::SYSTEM, system, configure))
        {
          mManagedByEngine.push_back(T::type::SYSTEM);
        }
        else
        {
          delete system;
          return 0;
        }
        return system;
      }

      /**
       * Add system to the engine
       *
       * @param name System name to associate with. Entity factory uses name to select proper system
       * @param configure system after add
       * @param system System that handles components
       * @returns true if succeed
       */
      bool addSystem(const std::string& name, EngineSystem* system, bool configure = true);

      /**
       * Configure system (call initialize and configure methods)
       *
       * @param name system name to configure
       * @returns true if succeed
       */
      bool configureSystem(const std::string& name);

      /**
       * Configure system (call initialize and configure methods)
       *
       * @param name system name to configure
       * @param restart force system restart
       * @returns true if succeed
       */
      bool configureSystem(const std::string& name, bool restart);

      /**
       * Configure system (call initialize method)
       *
       * @param name system name to configure
       * @param config new system config
       * @param restart should restart
       * @returns true if succeed
       */
      bool configureSystem(const std::string& name, const DataProxy& config, bool restart);

      /**
       * Check that engine has specified system
       *
       * @param name System name
       */
      bool hasSystem(const std::string& name) { return mEngineSystems.count(name) != 0; }
      /**
       * Get system by name
       * @param name System name
       */
      EngineSystem* getSystem(const std::string& name);
      /**
       * Get system by name
       */
      template<typename C>
      C* getSystem()
      {
        EngineSystem* system = getSystem(C::type::SYSTEM);
        if(!system)
          return 0;

        std::string type = system->getSystemInfo().get("type", "");
        if(!type.empty() && type != C::ID) {
          return 0;
        }

        return static_cast<C*>(system);
      }
      /**
       * Remove system by name
       * @param name System name
       * @returns true if system was deleted, false otherwise
       */
      bool removeSystem(const std::string& name);
      /**
       * Remove all systems, that were create by the engine
       * Remove all active systems
       */
      void removeSystems();
      /**
       * Gets all engine systems
       */
      EngineSystems& getSystems();
      /**
       * Create entity from the entity data
       * @param data Deserialized data object
       */
      Entity* createEntity(DataProxy& data);
      /**
       * Remove entity by id
       *
       * @param id entity id
       */
      bool removeEntity(const std::string& id);
      /**
       * Remove entity by pointer
       *
       * @param entity Entity pointer
       */
      bool removeEntity(Entity* entity);
      /**
       * Destroy all entities
       */
      void unloadAll();
      /**
       * Destroy entities that match filter
       *
       * @param f Matcher function
       */
      void unloadMatching(UnloadMatcherFunc f);
      /**
       * Get component by entity and name
       * @param entity Entity with all components
       * @param name Component name
       *
       * @deprecated
       */
      template<typename C> C* getComponent(Entity& entity, const std::string& name) { return static_cast<C*>(entity[name]); }
      /**
       * Get component by entity and name
       * @param entity Entity with all components
       * @param name Component name
       *
       * @deprecated
       */
      template<typename C> C* getComponent(Entity* entity, const std::string& name) { return getComponent<C>(*entity, name); }
      /**
       * Get component by entity and name
       * @param entity Entity with all components
       * @param name Component name
       */
      template<typename C> C* getComponent(Entity& entity) { return static_cast<C*>(entity[C::SYSTEM]); }
      /**
       * Get component by entity
       * @param entity Entity with all components
       * @param name Component name
       */
      template<typename C> C* getComponent(Entity* entity) { return getComponent<C>(*entity); }
      /**
       * Get component by entity
       * @param entity Entity with all components
       * @param name Component name
       */
      template<typename C> C* getComponent(const std::string& name)
      {
        Entity* e = getEntity(name);
        if(!e)
          return 0;

        return static_cast<C*>((*e)[C::SYSTEM]);
      }
      /**
       * Get component by entity
       * @param entity Entity with all components
       * @param name Component name
       */
      EntityComponent* getComponent(Entity* e, const std::string& name)
      {
        return (*e)[name];
      }
      /**
       * Get entity list
       */
      ObjectPool<Entity>::PointerVector getEntities() { return mEntities.getElements(); };
      /**
       * Get entity by id
       * @param id Entity id
       */
      Entity* getEntity(const std::string& id);

      /**
       * Get environment
       */
      inline const DataProxy& env() const { return mEnvironment; }

      /**
       * Get settings
       */
      inline const DataProxy& settings() const { return mConfiguration; }

      /**
       * Update global configuration property
       *
       * @param key Config key
       * @param value Config value
       */
      template<class T>
      void setEnv(const std::string& key, T value) {
        mEnvironment.put(key, value);
        fireEvent(EngineEvent(EngineEvent::ENV_UPDATED));
      }

      /**
       * Shutdown the engine
       * @param terminate Abort all running threads
       */
      void shutdown(bool terminate = false);

      /**
       * Callback to call in the main thread
       */
      typedef std::function<void()> MainThreadCallback;

      /**
       * Return value from main thread callback queueing
       */
      struct QueuedCallback {
        QueuedCallback(MainThreadCallback func) : func(func) {
          channel = std::make_shared<SignalChannel>();
        }
        MainThreadCallback func;
        std::shared_ptr<SignalChannel> channel;

        void wait() {
          channel->recv();
        }

        void execute()
        {
          func();
          channel->send(ChannelSignal::DONE);
        }
        private:
          QueuedCallback(const QueuedCallback&);
      };

      typedef std::shared_ptr<QueuedCallback> QueuedCallbackPtr;

      /**
       * Queue callback for execution in the main thread
       *
       * @param callback Callback function
       * @returns QueuedCallback which can be used to make call blocking
       */
      QueuedCallbackPtr executeInMainThread(MainThreadCallback func);

    private:
      class SystemWorker
      {
        public:
          SystemWorker(EngineSystem* system);
          virtual ~SystemWorker();

          /**
           * Start the worker
           */
          void start();

          /**
           * Stops the worker
           */
          void shutdown(bool terminate);

        private:
          friend class Engine;

          /**
           * Run loop
           */
          void run();
          EngineSystem* mSystem;
          std::thread* mThread;
          std::atomic_bool mShutdown;
          std::chrono::high_resolution_clock::time_point mPreviousUpdateTime;
          DataProxy mSystemConfig;
      };
      /**
       * Create component for entity
       *
       * @param entity Pointer to entity object
       * @param type System type to create into
       * @param node DataProxy with configs
       */
      bool createComponent(Entity* entity, const std::string& type, const DataProxy& node);
      /**
       * Update entity with new config
       *
       * @param entity Entity to update
       * @param node DataProxy with new config
       */
      bool readEntityData(Entity* entity, const DataProxy& node);

      /**
       * Shutdown threads
       *
       * @param terminate Abort execution
       */
      void shutdownThreads(bool terminate);

      bool mInitialized;

      EngineSystems mEngineSystems;
      Entities mEntities;
      unsigned long mEntityCounter;
      typedef std::map<const std::string, Entity*> EntityMap;
      EntityMap mEntityMap;

      typedef std::vector<std::string> SystemNames;
      SystemNames mSetUpOrder;
      SystemNames mManagedByEngine;

      DataProxy mConfiguration;
      DataProxy mEnvironment;

      typedef std::map<std::string, std::unique_ptr<SystemWorker>> Workers;
      Workers mWorkers;

      typedef ThreadSafeQueue<std::shared_ptr<QueuedCallback>> QueuedCallbacks;
      QueuedCallbacks mMainThreadCallbacks;
  };
}

#endif
